@@ -16,6 +16,8 @@ from typing import Any, Dict, Optional
 
 import attr
 
+from synapse.types import JsonDict
+
 from ._base import Config
 
 logger = logging.getLogger(__name__)
@@ -24,7 +26,7 @@ LEGACY_TEMPLATE_DIR_WARNING = """
 This server's configuration file is using the deprecated 'template_dir' setting in the
 'sso' section. Support for this setting has been deprecated and will be removed in a
 future version of Synapse. Server admins should instead use the new
-'custom_templates_directory' setting documented here:
+'custom_template_directory' setting documented here:
 https://matrix-org.github.io/synapse/latest/templates.html
 ---------------------------------------------------------------------------------------"""
 
@@ -49,7 +51,7 @@ class SSOConfig(Config):
 
     section = "sso"
 
-    def read_config(self, config, **kwargs) -> None:
+    def read_config(self, config: JsonDict, **kwargs: Any) -> None:
         sso_config: Dict[str, Any] = config.get("sso") or {}
 
         # The sso-specific template_dir
@@ -105,43 +107,3 @@ class SSOConfig(Config):
             self.root.server.public_baseurl + "_matrix/static/client/login"
         )
         self.sso_client_whitelist.append(login_fallback_url)
-
-    def generate_config_section(self, **kwargs) -> str:
-        return """\
-        # Additional settings to use with single-sign on systems such as OpenID Connect,
-        # SAML2 and CAS.
-        #
-        # Server admins can configure custom templates for pages related to SSO. See
-        # https://matrix-org.github.io/synapse/latest/templates.html for more information.
-        #
-        sso:
-            # A list of client URLs which are whitelisted so that the user does not
-            # have to confirm giving access to their account to the URL. Any client
-            # whose URL starts with an entry in the following list will not be subject
-            # to an additional confirmation step after the SSO login is completed.
-            #
-            # WARNING: An entry such as "https://my.client" is insecure, because it
-            # will also match "https://my.client.evil.site", exposing your users to
-            # phishing attacks from evil.site. To avoid this, include a slash after the
-            # hostname: "https://my.client/".
-            #
-            # The login fallback page (used by clients that don't natively support the
-            # required login flows) is whitelisted in addition to any URLs in this list.
-            #
-            # By default, this list contains only the login fallback page.
-            #
-            #client_whitelist:
-            #  - https://riot.im/develop
-            #  - https://my.custom.client/
-
-            # Uncomment to keep a user's profile fields in sync with information from
-            # the identity provider. Currently only syncing the displayname is
-            # supported. Fields are checked on every SSO login, and are updated
-            # if necessary.
-            #
-            # Note that enabling this option will override user profile information,
-            # regardless of whether users have opted-out of syncing that
-            # information when first signing in. Defaults to false.
-            #
-            #update_profile_information: true
-        """
